// 'Main.ts' file
import {Individual, Kid} from "./Abstract_optional.js"; // "./Abstract_optional" for compilation...

const main = () => { // 'window.onload = main;'
    const chromosomes_: ReadonlyArray<string> = ["Genetic female - XX", "Genetic male - XY"];

    // chromosomes_.push("YY"); // Compilation error, content cannot be changed...
    /* function switch_first_and_last<T>(array: readonly T[]): void {
        console.assert(array && array.length > 0);
        const temp: T = array[0];
        array[0] = array[array.length - 1]; // Compilation error because of 'readonly'...
        array[array.length - 1] = temp; // Compilation error because of 'readonly'...
    } */
    // Immutable tuple:
    const point: readonly [number, number] = [0, 0];

    let my_var: any;
    window.console.assert(my_var === undefined, "Note that 'undefined' is both the type and its singleton value!");
    my_var = {given_name: "Franck"};
    window.console.assert(my_var.given_name === "Franck", "No compilation error...");

    let my_other_var: unknown;
    my_other_var = {given_name: "Franck"};
    // window.console.assert(my_other_var.given_name === "Franck", "Compilation error!");
    // if ('given_name' in my_other_var) // Compilation error!
    //     window.alert(my_other_var.given_name); // Compilation error!

    function owns_given_name(o: any): o is { given_name: string } { // Type guard is introduced...
        return 'given_name' in o;
    }

    if (owns_given_name(my_other_var))
        window.alert(my_other_var.given_name); // No compilation error...


    // const Medal = Object.freeze({
    //     Gold: Symbol("Gold"),
    //     Silver: Symbol("Silver"),
    //     Bronze: Symbol("Bronze")
    // });
    // let award = Medal.Gold;
    // window.alert(award !== Symbol("Gold")); // 'true' because 'Symbol' objects are unique...

    // enum Medal {
    //     Gold = "Gold",
    //     Silver = "Silver",
    //     Bronze = "Bronze"
    // }
    // let award: Medal = Medal.Gold;
    // window.alert(award); // 'Gold'

    let matrix: Array<Array<number>> = [[1, 2], [1, 2]];

    const chromosomes: Array<string> = ["Genetic female - XX", "Genetic male - XY"];
    window.alert(chromosomes instanceof Array); // 'true' is displayed
    // window.alert(typeof chromosomes); // 'object' is displayed
    chromosomes[-1] = "N'importe quoi !";
    // window.alert(chromosomes[-1]); // "N'importe quoi !" is displayed
    window.alert("chromosomes.length: " + chromosomes.length); // "2" is displayed!

    const diseases = new Array(Symbol("Turner's syndrome - X"), Symbol("Klinefelter syndrome - XXY")); // Type inference...
    window.console.assert(diseases.length === 2);

    const genetics = [...chromosomes, ...diseases]; // Array of 'string | symbol'...
    // Symbols do not duplicate (they're unique!), so 'null':
    window.alert("'JSON.stringify(genetics)': " + JSON.stringify(genetics)); // '["Genetic female - XX","Genetic male - XY",null,null]' is displayed!
    const mapping: [string | symbol, string | symbol] = [chromosomes[0], diseases[0]]; // Tuple...

    /**
     * JavaScript 'number' primitive type is 64-bit long based on the IEEE 754 standard
     */
        // See also: http://speakingjs.com/es5/ch11.html#number_representation
    let one: number = Number.parseInt('00000001', 2);
    // if (Number.isNaN(one)) window.alert('Number.isNaN(one)');
    // else window.alert('one: ' + one); // '1'
    let two: Number = new Number(1 << 1);
    // window.alert('two(2): ' + two.toString(2)); // '10'
    let zero: Number = new Number(1 >> 1);
    // window.alert('zero: ' + zero); // '0'

    /**
     * JavaScript Bitwise operators apply on the first significant 32 bits only!
     */
        // How negative numbers are represented:
    let minus_nine: Number = new Number((~9) | 1); // '00000000000000000000000000001001' -> '11111111111111111111111111110110' -> // '11111111111111111111111111110111'
    window.alert('minus_nine: ' + minus_nine); // '-9'

    // 32 significant bits representing the number are viewed as an *unsigned* 32-bit integer:
    let x: Number = new Number(-1 >>> 0); // Unsigned right shift that inserts '0' on the left (BTW: unsigned left shift does not exist!)
    window.alert('-1 >>> 0: ' + x.toString(2)); // '11111111111111111111111111111111' -> (2^32 - 1) -> 4294967295

    window.alert(Number.MAX_SAFE_INTEGER.toString(2).length); // '53' since 'Number.MAX_SAFE_INTEGER' is equal to (2^53 - 1)
    window.console.assert(Number.MAX_SAFE_INTEGER === -Number.MIN_SAFE_INTEGER);

    window.alert(Number.MAX_VALUE * 2 === Number.POSITIVE_INFINITY);

    window.alert(isNaN(Math.sqrt(-1))); // 'true'

    let Jules: Individual = new Kid();
    window.alert(Jules.responsibilities()); // Unfortunately, bug...

    let is_Jules_responsible_: ReturnType<typeof Jules.isResponsible> = Jules.isResponsible(null);
    type Kid_bis_repetita = ThisParameterType<typeof Jules.isResponsible>;
    let Julie: Kid_bis_repetita;

    const F = "Franck"; // 'F' has the type "Franck"
    let B = "Barbier";   // 'B' has the type 'string'
    let FB = "FranckBarbier" as const; // 'FB' has type "FranckBarbier"
    // FB = "?"; // Compilation error because of 'as const'... -> somehow stupid as this stage?

    const lasagna = {country: "Italy"};
    lasagna.country = "France"; // 'const' has low scope...
    const foie_gras = {country: "France"} as const;
    // foie_gras.country = "Italy"; // Compilation error!
    // let my_favorite_dish = lasagna as const // 'const' cast only applies to references to enum members, string, number, boolean, array, or object literals.
}

main();

